import shutil
import json
import io
import os
import re
import platform
import sys

import semver

from .about import __version__


class InvalidPathPartsException(Exception): pass
class InvalidAppNameException(Exception): pass
class InvalidConstraintException(Exception): pass
class UnknownAppNameException(Exception): pass


def default_data_path(app_name):
    if app_name is None:
        raise InvalidAppNameException(app_name)
    try:
        mod = __import__(app_name, globals(), locals(), [], 0)
    except ImportError:
        raise UnknownAppNameException(app_name)
    return os.path.abspath(os.path.join(os.path.dirname(mod.__file__), 'data'))


def split_package_string(string):
    if not string:
        string = ''
    m = re.search(r'[^a-z0-9_]', string)
    if not m:
        return [string, '']
    return [string[:m.start()], string[m.start():].strip()]


def constraint_match(constraint_string, version):
    if not constraint_string:
        return True

    constraints = [c.strip() for c in constraint_string.split(',') if c.strip()]

    for c in constraints:
        if not re.match(r'[><=][=]?\d+(\.\d+)*', c):
            raise InvalidConstraintException('invalid constraint: %s' % c)

    return all(semver.match(version, c) for c in constraints)


def get_path(*path_parts, **kwargs):
    sep = kwargs.pop('sep', os.path.sep)
    if any(p for p in path_parts if '/' in p or '\\' in p):
        raise InvalidPathPartsException('avoid / and \\ in path parts: %s' % path_parts)
    return sep.join(path_parts)


def user_agent(name, version):
    uname = platform.uname()
    user_agent_vars = [
        ('Sputnik', __version__),
        (name, version),
        (platform.python_implementation(), platform.python_version()),
        (platform.uname()[0], uname[2]),
        ('64bits', sys.maxsize > 2**32)]

    return ' '.join(['%s/%s' % (k, v) for k, v in user_agent_vars if k])


def system_info(name, version):
    return {'app_name': name,
            'app_version': version,
            'sputnik_version': __version__,
            'py': platform.python_implementation(),
            'py_version': platform.python_version(),
            'os': platform.uname()[0],
            'os_version': platform.uname()[2],
            'bits': 64 if sys.maxsize > 2**32 else 32}


def expand_path(path):
    return path and os.path.expanduser(path) or path


def is_enough_space(path, size):
    if hasattr(shutil, "disk_usage"):  # python >= 3.3
        free = shutil.disk_usage(path).free
        return free >= size
    return True


def archive_filename(name, version, suffix=False):
    res = "%s-%s" % (name, version)
    if suffix:
        res += ".sputnik"
    return res


def json_dump(obj):
    defaults = {'sort_keys': True, 'indent': 2, 'separators': (',', ': ')}
    return json.dumps(obj, **defaults).encode('utf8')


def json_load(path):
    with io.open(path, "rb") as f:
        return json.loads(f.read().decode('utf8'))


def json_print(obj):
    defaults = {'sort_keys': True, 'indent': 2, 'separators': (',', ': ')}
    sys.stdout.write(json.dumps(obj, **defaults) + '\n')


def makedirs(path):
    path = os.path.dirname(path)
    if path and not os.path.exists(path):
        os.makedirs(path)


def unquote(s):
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        return s[1:-1]
    return s


def s3_header(value):
    if not re.match(r'[A-Za-z0-9-]+', value):
        raise Exception('invalid value for S3 header: %s' % value)
    return 'x-amz-meta-%s' % value.lower()


def dirpath(path):
    if os.path.isdir(path):
        return path
    return os.path.split(path)[0]


def dirname(path):
    return os.path.basename(dir(path))



import os
import logging

from . import util
from . import default
from .package_stub import PackageStub


class CachedPackage(PackageStub):
    def __init__(self, path):
        meta = util.json_load(os.path.join(path, default.META_FILENAME))
        super(CachedPackage, self).__init__(defaults=meta['package'])

        self.logger = logging.getLogger(__name__)
        self.meta = meta
        self.path = path

    @property
    def manifest(self):
        return self.meta['manifest']



import os
import codecs
import json
try:
    from http.cookiejar import MozillaCookieJar
except ImportError:
    from cookielib import MozillaCookieJar
try:
    from urllib.request import Request, build_opener, HTTPRedirectHandler, HTTPCookieProcessor
except ImportError:
    from urllib2 import Request, build_opener, HTTPRedirectHandler, HTTPCookieProcessor

from . import default
from . import util


class Session(object):
    def __init__(self, app_name, app_version, data_path, **kwargs):
        self.app_name = app_name
        self.app_version = app_version

        if not data_path or not os.path.isdir(data_path):
            raise Exception('invalid data_path: %s' % data_path)

        self.cookie_jar = MozillaCookieJar(os.path.join(data_path, default.COOKIES_FILENAME))
        try:
            self.cookie_jar.load()
        except EnvironmentError:
            pass

        self.opener = build_opener(
            HTTPRedirectHandler(),
            HTTPCookieProcessor(self.cookie_jar))

        super(Session, self).__init__(**kwargs)

    def open(self, request, default_charset=None):
        request.add_header('User-Agent', util.user_agent(self.app_name, self.app_version))

        system_string = json.dumps(util.system_info(self.app_name, self.app_version))
        request.add_header('X-Sputnik-System', system_string)

        r = self.opener.open(request)

        if hasattr(r.headers, 'get_content_charset'):  # py3
            charset = r.headers.get_content_charset() or default_charset
        elif hasattr(r.headers, 'getparam'):  # py2
            charset = r.headers.getparam('charset') or default_charset
        else:
            charset = default_charset

        if charset is None:
            return r
        return codecs.getreader(charset)(r)

    def __del__(self):
        if hasattr(self, 'cookie_jar'):
            self.cookie_jar.save()


class GetRequest(Request):
    pass


class HeadRequest(Request):
    def get_method(self):
        return 'HEAD'


class PutRequest(Request):
    def get_method(self):
        return 'PUT'



import os
import tarfile
import hashlib
import io

import pytest

from ..archive_writer import ArchiveWriter, InvalidPathException,\
                             EmptyArchiveException
from ..archive_reader import ArchiveReader
from ..default import ARCHIVE_FILENAME, META_FILENAME


def path_content(path, base_path=None):
    if base_path is None:
        base_path = path
    for root, _, filenames in os.walk(path):
        for filename in filenames:
            file_path = os.path.join(root, filename)
            yield tuple(os.path.relpath(file_path, base_path).split(os.path.sep))


def test_create_and_check(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path, base_path=sample_package_path)
    f.add_path(sample_package_path)
    f.close()

    assert os.path.exists(archive_path)

    tar = tarfile.open(archive_path, 'r')

    assert set(tar.getnames()) == set([ARCHIVE_FILENAME, META_FILENAME])


def test_create_and_read(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path, base_path=sample_package_path)
    f.add_path(sample_package_path)
    f.close()

    assert os.path.exists(archive_path)

    with ArchiveReader(archive_path) as f:
        assert set(f.list()) == set(path_content(sample_package_path))


def test_create_and_checksum(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path, base_path=sample_package_path)
    f.add_path(sample_package_path)
    f.close()

    assert os.path.exists(archive_path)

    with ArchiveReader(archive_path) as archive:
        for entry in archive.meta['manifest']:
            with io.open(os.path.join(sample_package_path, os.path.sep.join(entry['path'])), 'rb') as f:
                assert entry['checksum'][0] == hashlib.md5().name
                assert entry['checksum'][1] == hashlib.md5(f.read()).hexdigest()


def test_create_and_extract(tmp_path, tmp_path2, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path, base_path=sample_package_path)
    f.add_path(sample_package_path)
    f.close()

    assert os.path.exists(archive_path)

    with ArchiveReader(archive_path) as archive:
        archive.extract_all(tmp_path2)

    for root, _, filenames in os.walk(sample_package_path):
        for filename in filenames:
            path = os.path.relpath(os.path.join(root, filename), sample_package_path)
            package_path = os.path.join(sample_package_path, path)
            extract_path = os.path.join(tmp_path2, path)

            assert os.stat(package_path).st_size == os.stat(extract_path).st_size

            assert hashlib.md5(io.open(package_path, 'rb').read()).hexdigest() == \
                   hashlib.md5(io.open(extract_path, 'rb').read()).hexdigest()


def test_create_with_index(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path, base_path=sample_package_path)
    f.add_path(os.path.join(sample_package_path, 'data'))
    f.close()

    assert os.path.exists(archive_path)

    tar = tarfile.open(archive_path, 'r')

    assert set(tar.getnames()) == set([ARCHIVE_FILENAME, META_FILENAME])

    with ArchiveReader(archive_path) as f:
        content = path_content(os.path.join(sample_package_path, 'data'),
                               base_path=sample_package_path)
        assert set(f.list()) == set(content)


def test_create_without_base_path(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    os.chdir(sample_package_path)

    f = ArchiveWriter(archive_path)
    f.add_path('data')
    f.close()

    assert os.path.exists(archive_path)

    tar = tarfile.open(archive_path, 'r')

    assert set(tar.getnames()) == set([ARCHIVE_FILENAME, META_FILENAME])

    with ArchiveReader(archive_path) as f:
        content = path_content(os.path.join(sample_package_path, 'data'),
                               base_path=sample_package_path)
        assert set(f.list()) == set(content)


def test_create_abspath_without_base_path(tmp_path, sample_package_path):
    archive_path = os.path.join(tmp_path, 'test.sputnik')

    f = ArchiveWriter(archive_path)

    with pytest.raises(InvalidPathException):
        f.add_path(sample_package_path)

    with pytest.raises(EmptyArchiveException):
        f.close()

    assert not os.path.exists(f.tmp_path)
    assert not os.path.exists(archive_path)
    assert os.path.exists(tmp_path)
    assert len(os.listdir(tmp_path)) == 0
